from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import jijbench.exceptions.exceptions as exceptions
from jijbench.exceptions.exceptions import ConcurrentFailedError, SolverFailedError

__all__ = ["exceptions", "ConcurrentFailedError", "SolverFailedError"]



from __future__ import annotations

import json
import os
import pathlib
from typing import Dict, List, Tuple, Union

import jijmodeling as jm

from jijbench import datasets

__all__ = []


class JijModelingTarget:
    def __init__(
        self, problem, instance: Union[Tuple[str, Dict], List[Tuple[str, Dict]]]
    ):
        self.problem = problem
        self.instance = instance


class DefaultInstanceMixin:
    def _instance_dir(self, size: str):
        instance_dir = (
            pathlib.Path(__file__)
            .parent.joinpath("Instances")
            .joinpath(size)
            .joinpath(self.problem_name)
        )
        return instance_dir

    def instance_names(self, size: str) -> List[str]:
        instance_dir = self._instance_dir(size)
        instance_names = [
            p.name.split(".")[0]
            for p in instance_dir.glob(os.path.join("**", "*.json"))
        ]
        return instance_names

    def small_instance(self) -> List[Tuple[str, Dict]]:
        return [
            (name, self.load("small", name)) for name in self.instance_names("small")
        ]

    def medium_instance(self) -> List[str]:
        return [
            (name, self.load("medium", name)) for name in self.instance_names("medium")
        ]

    def large_instance(self) -> List[str]:
        return [
            (name, self.load("large", name)) for name in self.instance_names("large")
        ]

    def get_instance(self, size: str, instance_name: str) -> Dict:
        return self.load(size, instance_name=instance_name)

    def load(self, size: str, instance_name: str):
        instance_file_path = None
        for file_path in self._instance_dir(size).glob(os.path.join("**", "*.json")):
            if file_path.name.split(".")[0] == instance_name:
                instance_file_path = file_path
        if instance_file_path is None:
            raise FileNotFoundError(
                f"'{size}/{self.problem_name}/{instance_name}.json' is not found."
            )

        instance_path = instance_file_path.resolve()
        with open(instance_path, "r") as f:
            instance_data = json.load(f)

        return instance_data


class Knapsack(JijModelingTarget, DefaultInstanceMixin):
    def _problem(problem_name):
        w = jm.Placeholder("w", dim=1)
        v = jm.Placeholder("v", dim=1)
        n = jm.Placeholder("n")
        c = jm.Placeholder("c")
        x = jm.Binary("x", shape=(n,))

        # i: itemの添字
        i = jm.Element("i", n)

        problem = jm.Problem(problem_name)

        # objective function
        obj = jm.Sum(i, v[i] * x[i])
        problem += -1 * obj

        # Constraint: knapsack 制約
        const = jm.Constraint("knapsack-constraint", jm.Sum(i, w[i] * x[i]) - c <= 0)
        problem += const

        return problem

    problem_name = "knapsack"
    problem = _problem(problem_name)

    def __init__(self):
        super().__init__(self.problem, self.small_instance()[0])


class BinPacking(JijModelingTarget, DefaultInstanceMixin):
    def _problem(problem_name):
        w = jm.Placeholder("w", dim=1)
        num_items = jm.Placeholder("n")
        c = jm.Placeholder("c")

        # y[j]: bin j を使用するかしないか
        y = jm.Binary("y", shape=(num_items,))
        # x[i][j]: item i を bin j に入れるとき1
        x = jm.Binary("x", shape=(num_items, num_items))

        # i: itemの添字
        i = jm.Element("i", num_items)
        # j: binの添字
        j = jm.Element("j", num_items)

        problem = jm.Problem(problem_name)

        # objective function
        obj = y[:]
        problem += obj

        # Constraint1: 各itemをちょうど1つのbinにぶち込む
        const1 = jm.Constraint(
            "onehot-constraint", jm.Sum(j, x[i, j]) - 1 == 0, forall=i
        )
        problem += const1

        # Constraint2: knapsack制約
        const2 = jm.Constraint(
            "knapsack-constraint", jm.Sum(i, w[i] * x[i, j]) - y[j] * c <= 0, forall=j
        )
        problem += const2

        return problem

    problem_name = "bin-packing"
    problem = _problem(problem_name)

    def __init__(self):
        super().__init__(self.problem, self.small_instance()[0])


class NurseScheduling(JijModelingTarget, DefaultInstanceMixin):
    def _problem(problem_name):
        # 問題
        problem = jm.Problem(problem_name)

        I = jm.Placeholder("I")  # 人の数
        D = jm.Placeholder("D")  # 日数 D%7=0 とし, 月曜日からstartするとする
        W = jm.Placeholder("W")  # 週の数
        T = jm.Placeholder("T")  # シフトのタイプ数

        N = jm.Placeholder("N", dim=2)  # N[i] = 人iが必ず休む日

        # シフトtの次の日に働ないtypeのシフト R = [type1, type2, type3, ... ] = [[1, 2], [0, ], [], ...]
        R = jm.Placeholder("R", dim=2)
        l = jm.Placeholder("l", shape=(I,))  # シフトtの労働時間
        m_max = jm.Placeholder("m_max", shape=(I, T)).set_latex(
            "\mathrm{m\_max}"
        )  # 従業員にシフトtを割当てられる最大回数
        b_min = jm.Placeholder("b_min", shape=(I,)).set_latex(
            "\mathrm{b\_min}"
        )  # 各従業員の労働時間の最小値
        b_max = jm.Placeholder("b_max", shape=(I,)).set_latex(
            "\mathrm{b\_max}"
        )  # 各従業員の労働時間の最大値
        c_min = jm.Placeholder("c_min", shape=(I,)).set_latex(
            "\mathrm{c\_min}"
        )  # 各従業員の最小連続勤務数
        c_max = jm.Placeholder("c_max", shape=(I,)).set_latex(
            "\mathrm{c\_max}"
        )  # 各従業員の最大連続勤務数
        o_min = jm.Placeholder("o_min", shape=(I,)).set_latex(
            "\mathrm{o\_min}"
        )  # 各従業員の連続最低休日日数
        a_max = jm.Placeholder("a_max", shape=(I,)).set_latex(
            "\mathrm{a\_max}"
        )  # 各従業員の週末働ける最大の回数
        # q[i, d, t] = 1, 人iは日にちdにtype tの仕事をしたい
        q = jm.Placeholder("q", shape=(I, D, T))
        # p[i, d, t] = 1, 人iは日にちdにtype tの仕事をしたくない
        p = jm.Placeholder("p", shape=(I, D, T))
        u = jm.Placeholder("u", shape=(D, T))  # 日にちd, type tの必要人数
        v_min = jm.Placeholder("v_min", shape=(D, T)).set_latex(
            "\mathrm{v\_min}"
        )  # 人員不足のペナルティーの重み
        v_max = jm.Placeholder("v_max", shape=(D, T)).set_latex(
            "\mathrm{v\_max}"
        )  # 人員過剰のペナルティーの重み

        len_R = R.shape[0]

        # Element
        i = jm.Element("i", (0, I))
        d = jm.Element("d", (0, D))
        t = jm.Element("t", (0, T))
        w = jm.Element("w", (0, W))
        dt = jm.Element("dt", (0, len_R))
        j = jm.Element("j", (0, D))
        s = jm.Element("s", (0, D))

        # 決定変数
        x = jm.Binary("x", shape=(I, D, T))  # 人iを日にちdにシフトtを割当てるかどうか
        k = jm.Binary("k", shape=(I, W))  # week wに働いたかどうか
        y = jm.Integer("y", lower=0, upper=u, shape=(D, T))  # 日dのシフトtの不足人数
        z = jm.Integer("z", lower=0, upper=u, shape=(D, T))  # 日dのシフトtの過剰人数

        # Objective Function
        term1 = jm.Sum([i, d, t], q[i, d, t] * (1 - x[i, d, t]))  # シフト入りの希望を叶える
        term2 = jm.Sum([i, d, t], p[i, d, t] * x[i, d, t])  # シフト休みの希望を叶える
        term3 = jm.Sum([d, t], y[d, t] * v_min[d, t])  # 人員不足のペナルティー
        term4 = jm.Sum([d, t], z[d, t] * v_max[d, t])  # 人員過剰のペナルティー
        problem += term1 + term2 + term3 + term4

        # Constraint1: 1人一つの仕事しか割り当てられない
        const1 = jm.Sum(t, x[i, d, t])
        problem += jm.Constraint("assign", const1 - 1 <= 0, forall=[i, d])

        # Constraint2: 特定の仕事を行った次の日に, 特定の仕事を行うことはできない
        problem += jm.Constraint(
            "shift-rotation",
            x[i, d, t] + x[i, (d + 1), dt] - 1 <= 0,
            forall=[i, (d, (d != D - 1)), t, (dt, (R[t, dt] != -1))],
        )

        # Constraint3: 従業員に割り当てられる各タイプのシフトの最大数
        const3 = jm.Sum(d, x[i, d, t]) - m_max[i, t]
        problem += jm.Constraint("assign-max", const3 <= 0, forall=[i, t])

        # Constraint4: 最低・最高労働時間
        const4 = jm.Sum([d, t], l[t] * x[i, d, t])
        problem += jm.Constraint(
            "minimum-work-time",
            b_min[i] - const4 <= 0,
            forall=[
                i,
            ],
        )
        problem += jm.Constraint(
            "maximum-work-time",
            const4 - b_max[i] <= 0,
            forall=[
                i,
            ],
        )

        # Constraint5: 最大連続勤務
        j5 = jm.Element("j5", (d, d + c_max[i] + 1))
        const5 = jm.Sum([j5, t], x[i, j5, t])
        problem += jm.Constraint(
            "maximum-consecutive-shifts",
            const5 - c_max[i] <= 0,
            forall=[i, (d, d <= D - (c_max[i] + 1))],
        )

        # Constraint6: 最小連続勤務
        s6 = jm.Element("s6", (1, c_min[i]))
        j6 = jm.Element("j6", (d + 1, d + s6 + 1))
        problem += jm.Constraint(
            "minimum-consecutive-shifts",
            jm.Sum(t, x[i, d, t])
            + (s6 - jm.Sum([j6, t], x[i, j6, t]))
            + jm.Sum(t, x[i, d + s6 + 1, t])
            - 1
            >= 0,
            forall=[i, s6, (d, d < (D - (s6 + 1)))],
        )

        # Constraint7: 最低連続休暇日数
        s7 = jm.Element("s7", (1, o_min[i]))
        j7 = jm.Element("j7", (d + 1, d + s7 + 1))
        problem += jm.Constraint(
            "minimum-consecutive-days-off",
            (1 - jm.Sum(t, x[i, d, t]))
            + jm.Sum([j7, t], x[i, j7, t])
            + (1 - jm.Sum(t, x[i, d + s7 + 1, t]))
            - 1
            >= 0,
            forall=[i, s7, (d, d < (D - (s7 + 1)))],
        )

        # Constraint8: 週末休みの最大回数
        const8 = jm.Sum(t, x[i, 7 * (w + 1) - 2, t]) + jm.Sum(
            t, x[i, 7 * (w + 1) - 1, t]
        )
        problem += jm.Constraint(
            "variable-k-lower", k[i, w] - const8 <= 0, forall=[i, w]
        )
        problem += jm.Constraint(
            "variable-k-upper", const8 - 2 * k[i, w] <= 0, forall=[i, w]
        )
        problem += jm.Constraint(
            "maximum-number-of-weekends",
            jm.Sum(w, k[i, w]) - a_max[i] <= 0,
            forall=[
                i,
            ],
        )

        # Constraint9: 働けない日の制約
        d_o = jm.Element("do", N[i]).set_latex("{\mathrm d\_o}")
        problem += jm.Constraint(
            "days-off", x[i, d_o, t] == 0, forall=[i, (d_o, d_o != -1), t]
        )

        # Constraint10: 必要人数に関する制約
        const10 = jm.Sum(i, x[i, d, t]) - z[d, t] + y[d, t] - u[d, t]
        problem += jm.Constraint("cover-requirements", const10 == 0, forall=[d, t])

        return problem

    problem_name = "nurse-scheduling"
    problem = _problem(problem_name)

    def __init__(self):
        super().__init__(self.problem, self.small_instance()[0])


class TSPTW(JijModelingTarget, DefaultInstanceMixin):
    def _problem(problem_name):
        # 問題
        problem = jm.Problem(problem_name)

        # 距離行列
        dist = jm.Placeholder("d", dim=2)  # 距離行列
        N = jm.Placeholder("N")
        e = jm.Placeholder("e", shape=(N,))  # ready time
        l = jm.Placeholder("l", shape=(N,))  # due time
        x = jm.Binary("x", shape=(N, N))
        t = jm.Integer("t", shape=(N,), lower=e, upper=l)

        i = jm.Element("i", N)
        j = jm.Element("j", N)

        # Objevtive Function: 距離の最小化
        sum_list = [i, (j, j != i)]
        obj = jm.Sum(sum_list, dist[i, j] * x[i, j])
        problem += obj

        # Const1: 都市iから出る辺は1つ
        term1 = jm.Sum((j, j != i), x[i, j])
        problem += jm.Constraint(
            "onehot-constraint1",
            term1 == 1,
            forall=[
                i,
            ],
        )

        # Const2: 都市iに入る辺は1つ
        term2 = jm.Sum((j, j != i), x[j, i])
        problem += jm.Constraint(
            "onehot-constraint2",
            term2 == 1,
            forall=[
                i,
            ],
        )

        # Const3: Time Windows制約
        term3 = t[i] + dist[i, j] - t[j] - 20 * (1 - x[i, j])
        forall_list = [(j, j != 0), (i, (i != 0) & (i != j))]
        problem += jm.Constraint(
            "time-window-constraint", term3 <= 0, forall=forall_list
        )

        return problem

    problem_name = "travelling-salesman-with-time-windows"
    problem = _problem(problem_name)

    def __init__(self):
        super().__init__(self.problem, self.small_instance()[0])


class TSP(JijModelingTarget, DefaultInstanceMixin):
    def _problem(problem_name):
        # 問題
        problem = jm.Problem(problem_name)
        dist = jm.Placeholder("d", dim=2)
        N = jm.Placeholder("N")

        x = jm.Binary("x", shape=(N, N))
        i = jm.Element("i", N)
        j = jm.Element("j", N)

        t = jm.Element("t", N)

        # Objective Funtion
        sum_list = [t, i, j]
        obj = jm.Sum(sum_list, dist[i, j] * x[t, i] * x[(t + 1) % N, j])
        problem += obj

        # const1: onehot for time
        const1 = x[t, :]
        problem += jm.Constraint(
            "onehot-time",
            const1 == 1,
            forall=[
                t,
            ],
        )

        # const2: onehot for location
        const2 = x[:, i]
        problem += jm.Constraint(
            "onehot-location",
            const2 == 1,
            forall=[
                i,
            ],
        )

        return problem

    problem_name = "travelling-salesman"
    problem = _problem(problem_name)

    def __init__(self):
        super().__init__(self.problem, self.small_instance()[0])


def get_problem(problem_name):
    return getattr(datasets, problem_name).problem



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import jijbench.functions as functions
import jijbench.node as node
from jijbench.benchmark.benchmark import Benchmark, construct_benchmark_for
from jijbench.containers.containers import Artifact, Record, Table
from jijbench.datasets.instance_data import get_instance_data
from jijbench.datasets.problem import get_problem
from jijbench.elements.array import Array
from jijbench.elements.base import Callable, Number, String
from jijbench.elements.date import Date
from jijbench.elements.id import ID
from jijbench.evaluation.evaluation import Evaluation
from jijbench.experiment.experiment import Experiment
from jijbench.io.io import load, save
from jijbench.solver.base import Parameter, Response, Solver
from jijbench.solver.jijzept import InstanceData, SampleSet, UserDefinedModel

__all__ = [
    "construct_benchmark_for",
    "functions",
    "node",
    "get_instance_data",
    "get_problem",
    "load",
    "save",
    "Array",
    "Artifact",
    "Benchmark",
    "Callable",
    "Date",
    "Evaluation",
    "Experiment",
    "ID",
    "InstanceData",
    "UserDefinedModel",
    "Number",
    "Parameter",
    "Record",
    "Response",
    "SampleSet",
    "Solver",
    "Table",
    "String",
]



from __future__ import annotations

import pandas as pd

from jijbench.containers.containers import Artifact, Table
from jijbench.experiment.experiment import Experiment
from jijbench.functions.concat import Concat
from jijbench.functions.factory import RecordFactory
from jijbench.functions.metrics import (
    FeasibleRate,
    ResidualEnergy,
    SuccessProbability,
    TimeToSolution,
)
from jijbench.node.base import FunctionNode
from jijbench.solver.jijzept import SampleSet


class Evaluation(FunctionNode[Experiment, Experiment]):
    """Evaluate the benchmark results."""

    def __call__(
        self,
        inputs: list[Experiment],
        opt_value: float | None = None,
        pr: float = 0.99,
    ) -> Experiment:
        return super().__call__(inputs, opt_value=opt_value, pr=pr)

    def operate(
        self,
        inputs: list[Experiment],
        opt_value: float,
        pr: float = 0.99,
    ) -> Experiment:
        """Calculate the typincal metrics of benchmark results.

        The metrics are as follows:
            - success_probability: Solution that is feasible and less than or equal to opt_value is counted as success, which is NaN if `opt_value` is not given.
            - feasible_rate: Rate of feasible solutions out of all solutions.
            - residual_energy: Difference between average objective of feasible solutions and `opt_value`, which is NaN if `opt_value` is not given.
            - TTS(optimal): Time to obtain opt_value with probability `pr`, which is NaN if opt_value is not given.
            - TTS(feasible): Time to obtain feasible solutions with probability `pr`.
            - TTS(derived): Time to obtain minimum objective among feasible solutions with probability `pr`.

        Args:
            opt_value (float, optional): Optimal value for instance_data.
            pr (float, optional): Probability of obtaining optimal value. Defaults to 0.99.

        Returns:
            Experiment: Experiment object included evalution results.
        """

        def f(x: pd.Series, opt_value: float, pr: float) -> pd.Series:
            inputs: list[SampleSet] = x.tolist()
            node = Concat()(inputs)
            metrics = [
                SuccessProbability()([node], opt_value=opt_value),
                FeasibleRate()([node]),
                ResidualEnergy()([node], opt_value=opt_value),
                TimeToSolution()([node], pr=pr, opt_value=opt_value, base="optimal"),
                TimeToSolution()([node], pr=pr, base="feasible"),
                TimeToSolution()([node], pr=pr, base="derived"),
            ]
            record = RecordFactory()(metrics)
            return record.data

        experiment = Concat()(inputs)

        artifact, table = experiment.data

        sampleset_columns = [
            col
            for col in table.columns
            if all(map(lambda x: isinstance(x, SampleSet), table.data[col]))
        ]

        data = table.data[sampleset_columns].apply(
            f, opt_value=opt_value, pr=pr, axis=1
        )
        metrics_artifact = Artifact(data.to_dict("index"))
        artifact = Concat()([artifact, metrics_artifact])

        metrics_table = Table(data)
        table = Concat()([table, metrics_table], axis=1)

        experiment.data = (artifact, table)
        return experiment



from __future__ import annotations

import pathlib
import typing as tp

import jijmodeling as jm
import pandas as pd
from typing_extensions import TypeGuard

from jijbench.consts.path import DEFAULT_RESULT_DIR
from jijbench.elements.id import ID
from jijbench.node.base import DataNode, FunctionNode
from jijbench.typing import DataNodeT

if tp.TYPE_CHECKING:
    from jijbench.containers.containers import Artifact, Record, Table
    from jijbench.experiment.experiment import Experiment
    from jijbench.solver.jijzept import SampleSet


def _is_artifact_list(inputs: list[DataNodeT]) -> TypeGuard[list[Artifact]]:
    return all([node.__class__.__name__ == "Artifact" for node in inputs])


def _is_experiment_list(inputs: list[DataNodeT]) -> TypeGuard[list[Experiment]]:
    return all([node.__class__.__name__ == "Experiment" for node in inputs])


def _is_record_list(inputs: list[DataNodeT]) -> TypeGuard[list[Record]]:
    return all([node.__class__.__name__ == "Record" for node in inputs])


def _is_table_list(inputs: list[DataNodeT]) -> TypeGuard[list[Table]]:
    return all([node.__class__.__name__ == "Table" for node in inputs])


def _is_sampleset_list(inputs: list[DataNodeT]) -> TypeGuard[list[SampleSet]]:
    return all([node.__class__.__name__ == "SampleSet" for node in inputs])


def _is_datanode_list(inputs: list[DataNodeT]) -> bool:
    sample = inputs[0]
    is_datanode = isinstance(sample, DataNode)
    return all([isinstance(node, sample.__class__) for node in inputs]) & is_datanode


class Concat(FunctionNode[DataNodeT, DataNodeT]):
    """Concat class for concatenating multiple data nodes.
    This class can be apply to `Artifact`, `Experiment`, `Record`, `Table`, `SampleSet`.
    """

    @tp.overload
    def __call__(self, inputs: list[Artifact], name: tp.Hashable = None) -> Artifact:
        ...

    @tp.overload
    def __call__(
        self,
        inputs: list[Experiment],
        name: str | None = None,
        *,
        autosave: bool = True,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> Experiment:
        ...

    @tp.overload
    def __call__(self, inputs: list[Record], name: tp.Hashable = None) -> Record:
        ...

    @tp.overload
    def __call__(
        self,
        inputs: list[Table],
        name: tp.Hashable = None,
        *,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> Table:
        ...

    @tp.overload
    def __call__(self, inputs: list[SampleSet], name: str) -> SampleSet:
        ...

    def __call__(
        self,
        inputs: list[DataNodeT],
        name: tp.Hashable = None,
        *,
        autosave: bool = True,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> DataNode:
        """Concatenates the given list of mapping type objects.

        Args:
            inputs (DataNodeT): A list of artifacts, experiments, records, or tables. The type of elements in 'inputs' must be unified either 'Artifact', 'Experiment', 'Record' or 'Table'.
            name (tp.Hashable, optional): A name for the resulting data. Defaults to None.
            autosave (bool, optional): A flag indicating whether to save the result to disk. Defaults to True.
            savedir (str | pathlib.Path, optional): The directory to save the result in. Defaults to DEFAULT_RESULT_DIR.
            axis (tp.Literal[0, 1], optional): The axis to concatenate the inputs along. Defaults to 0.
            index_name (str | None, optional): The name of the index after concatenation. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' is not unified either 'Artifact', 'Experiment', 'Record' or 'Table'.

        Returns:
            DataNode: The resulting Artifact, Experiment, Record, Table or Sampleset object.
        """
        if _is_datanode_list(inputs):
            return super().__call__(
                inputs,
                name=name,
                autosave=autosave,
                savedir=savedir,
                axis=axis,
                index_name=index_name,
            )
        else:
            raise TypeError(
                "Type of elements in 'inputs' must be unified either 'Artifact', 'Experiment', 'Record' or 'Table'."
            )

    @tp.overload
    def operate(self, inputs: list[Artifact], name: tp.Hashable = None) -> Artifact:
        ...

    @tp.overload
    def operate(
        self,
        inputs: list[Experiment],
        name: str | None = None,
        *,
        autosave: bool = True,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    ) -> Experiment:
        ...

    @tp.overload
    def operate(self, inputs: list[Record], name: tp.Hashable = None) -> Record:
        ...

    @tp.overload
    def operate(
        self,
        inputs: list[Table],
        name: tp.Hashable = None,
        *,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> Table:
        ...

    @tp.overload
    def operate(self, inputs: list[SampleSet], name: str) -> SampleSet:
        ...

    def operate(
        self,
        inputs: list[DataNodeT],
        name: tp.Hashable = None,
        *,
        autosave: bool = True,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> DataNode:
        """This method operates the concatenation of the given 'inputs' either 'Artifact', 'Experiment', 'Record' or 'Table'
        objects into a single object of the same type as 'inputs'.

        Args:
            inputs (MappingListTypes): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.
            autosave (bool, optional): If True, the resulting object will be saved to disk. Defaults to True.
            savedir (str | pathlib.Path, optional): The directory where the resulting object will be saved. Defaults to DEFAULT_RESULT_DIR.
            axis (tp.Literal[0, 1], optional): The axis along which to concatenate the input 'Table' objects. Defaults to 0.
            index_name (str | None, optional): The name of the resulting object's index. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            MappingTypes: The resulting 'Artifact', 'Experiment', 'Record' or 'Table' object.

        """
        if _is_artifact_list(inputs):
            data = {}
            for node in inputs:
                for k, v in node.data.items():
                    if k in data:
                        data[k].update(v.copy())
                    else:
                        data[k] = v.copy()
            return type(inputs[0])(data, name)
        elif _is_experiment_list(inputs):
            concat_a: Concat[Artifact] = Concat()
            concat_t: Concat[Table] = Concat()
            inputs_a = [n.data[0] for n in inputs]
            inputs_t = [n.data[1] for n in inputs]
            artifact = inputs_a[0].apply(concat_a, inputs_a[1:])
            table = inputs_t[0].apply(
                concat_t, inputs_t[1:], axis=axis, index_name=index_name
            )

            if name is None:
                name = ID().data

            if not isinstance(name, str):
                raise TypeError("Attirbute name of Experiment must be string.")

            return type(inputs[0])(
                (artifact, table),
                name,
                autosave=autosave,
                savedir=savedir,
            )
        elif _is_record_list(inputs):
            data = pd.concat([node.data for node in inputs])
            return type(inputs[0])(data, name)
        elif _is_table_list(inputs):
            data = pd.concat([node.data for node in inputs], axis=axis)
            data.index.name = index_name
            return type(inputs[0])(data, name)
        elif _is_sampleset_list(inputs):
            data = jm.concatenate([node.data for node in inputs])
            return type(inputs[0])(data, str(name))
        else:
            raise TypeError(
                "Type of elements in 'inputs' must be unified either 'Artifact', 'Experiment', 'Record' or 'Table'."
            )



from __future__ import annotations

import typing as tp
import warnings

import jijmodeling as jm
import numpy as np

from jijbench.elements.base import Number
from jijbench.functions.concat import Concat
from jijbench.node.base import FunctionNode
from jijbench.solver.jijzept import SampleSet


def _is_success_list(sampleset: SampleSet, opt_value: int | float) -> list[bool]:
    is_feas = _is_feasible_list(sampleset)
    objective = np.array(sampleset.evaluation.objective)
    return list((objective <= opt_value) & is_feas)


def _is_feasible_list(sampleset: SampleSet) -> list[bool]:
    if sampleset.evaluation.constraint_violations is None:
        raise ValueError(
            "The value of sampleset.evaluation.constraint_violations is None. This SampleSet object is not evaluated or has not constraints."
        )
    else:
        constraint_violations = np.array(
            list(sampleset.evaluation.constraint_violations.values())
        )
        return constraint_violations.sum(axis=0) == 0


class Metrics(FunctionNode[SampleSet, Number]):
    """A base class for metrics."""

    def __call__(self, inputs: list[SampleSet], **kwargs: tp.Any) -> Number:
        """Calculate the metric.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
        """

        def _generate_warning_msg(metrics):
            return f'{self.__class__.__name__} cannot be calculated because "{metrics}" is not stored in table of Experiment object.'

        concat: Concat[SampleSet] = Concat()
        node = concat(inputs, "sampleset")

        objective = node.evaluation.objective
        if objective is None:
            raise ValueError("This SampleSet object is not evaluated.")
        if np.isnan(objective).any():
            warnings.warn(_generate_warning_msg("objective"))

        time = node.time.solve
        if time is None:
            warnings.warn(_generate_warning_msg("execution_time"))
        else:
            if time.solve is None:
                warnings.warn(_generate_warning_msg("execution_time"))

        num_occurrences = node.record.num_occurrences
        if np.isnan(num_occurrences).any():
            warnings.warn(_generate_warning_msg("num_occurrences"))

        return super().__call__([node], **kwargs)


class TimeToSolution(Metrics):
    """Time to solution.
    The time to solution is defined as the time to obtain a solution with probability `pr`.
    The solution is defined as the solution that is feasible and less than or equal to `opt_value`.
    There are three types of time to solution:
        - optimal: The solution is the optimal solution.
        - feasible: The solution is feasible.
        - derived: The solution is the best solution among the obtained solutions.
    """

    @tp.overload
    def __call__(
        self,
        inputs: list[SampleSet],
        pr: float,
        *,
        opt_value: int | float,
        base: tp.Literal["optimal"] = "optimal",
    ) -> Number:
        ...

    @tp.overload
    def __call__(
        self, inputs: list[SampleSet], pr: float, *, base: tp.Literal["feasible"]
    ) -> Number:
        ...

    @tp.overload
    def __call__(
        self, inputs: list[SampleSet], pr: float, *, base: tp.Literal["derived"]
    ) -> Number:
        ...

    def __call__(
        self,
        inputs: list[SampleSet],
        pr: float,
        *,
        opt_value: int | float | None = None,
        base: tp.Literal["optimal", "feasible", "derived"] = "optimal",
    ) -> Number:
        """Calculate time to solution.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
            pr (float): The probability of obtaining a solution.
            opt_value (int or float): The optimal value.
            base (str): The type of time to solution.

        Returns:
            Number: The time to solution.
        """
        return super().__call__(inputs, pr=pr, opt_value=opt_value, base=base)

    @tp.overload
    def operate(
        self,
        inputs: list[SampleSet],
        pr: float,
        *,
        opt_value: int | float,
        base: tp.Literal["optimal"] = "optimal",
    ) -> Number:
        ...

    @tp.overload
    def operate(
        self, inputs: list[SampleSet], pr: float, *, base: tp.Literal["feasible"]
    ) -> Number:
        ...

    @tp.overload
    def operate(
        self, inputs: list[SampleSet], pr: float, *, base: tp.Literal["derived"]
    ) -> Number:
        ...

    def operate(
        self,
        inputs: list[SampleSet],
        pr: float,
        *,
        opt_value: int | float | None = None,
        base: tp.Literal["optimal", "feasible", "derived"] = "optimal",
    ) -> Number:
        """Calculate time to solution.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
            pr (float): The probability of obtaining a solution.
            opt_value (int or float): The optimal value.
            base (str): The type of time to solution.

        Returns:
            Number: The time to solution.
        """

        node = inputs[0]

        num_occurrences = np.array(node.record.num_occurrences)
        if sum(num_occurrences) == 1:
            warnings.warn("num_reads = 1; should be increased to measure TTS")

        if base == "optimal":
            f_ps = SuccessProbability()
            ps = f_ps([node], opt_value=opt_value or np.nan)
        elif base == "feasible":
            f_ps = FeasibleRate()
            ps = f_ps([node])
        elif base == "derived":
            f_ps = SuccessProbability()
            is_feas = _is_feasible_list(node)
            if any(is_feas):
                opt_value = np.array(node.evaluation.objective)[is_feas].min()
                ps = f_ps([node], opt_value=opt_value or np.nan)
            else:
                ps = f_ps([node], opt_value=np.inf)

        time = np.nan if node.time.solve is None else node.time.solve.solve
        time = np.nan if time is None else time

        if ps.data == 1:
            data = 0.0
        elif ps:
            data = np.log(1 - pr) / np.log(1 - ps.data) * time
        else:
            data = np.inf
        return Number(data, f"TTS[{base}]")


class SuccessProbability(Metrics):
    """Success probability.
    The success probability is defined as the probability of obtaining a solution less than or equal to `opt_value`.
    """

    def __call__(self, inputs: list[SampleSet], opt_value: int | float) -> Number:
        """Calculate success probability.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
            opt_value (int or float): The optimal value.

        Returns:
            Number: The success probability.
        """
        return super().__call__(inputs, opt_value=opt_value)

    def operate(self, inputs: list[SampleSet], opt_value: int | float) -> Number:
        """Calculate success probability.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
            opt_value (int or float): The optimal value.

        Returns:
            Number: The success probability.
        """
        node = inputs[0]
        num_occurrences = np.array(node.record.num_occurrences)
        num_success = sum(_is_success_list(node, opt_value) * num_occurrences)
        data = num_success / sum(num_occurrences)
        return Number(data, "success_probability")


class FeasibleRate(Metrics):
    """Feasible rate.
    The feasible rate is defined as the probability of obtaining a feasible solution.
    """

    def __call__(self, inputs: list[SampleSet]) -> Number:
        """Calculate feasible rate.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.

        Returns:
            Number: The feasible rate.

        """
        return super().__call__(inputs)

    def operate(self, inputs: list[SampleSet]) -> Number:
        node = inputs[0]
        num_occurrences = np.array(node.record.num_occurrences)
        num_feasible = sum(_is_feasible_list(node) * num_occurrences)
        data = num_feasible / sum(num_occurrences)
        return Number(data, "feasible_rate")


class ResidualEnergy(Metrics):
    """Residual energy.
    The residual energy is defined as the mean of the objective function minus the optimal value.
    """

    def __call__(self, inputs: list[SampleSet], opt_value: int | float) -> Number:
        return super().__call__(inputs, opt_value=opt_value)

    def operate(self, inputs: list[SampleSet], opt_value: int | float) -> Number:
        node = inputs[0]

        is_feas = np.array(_is_feasible_list(node))
        if all(~is_feas):
            data = np.nan
        else:
            num_occurrences = np.array(node.record.num_occurrences)
            objective = np.array(node.evaluation.objective) * is_feas * num_occurrences
            mean = objective.sum() / (is_feas * num_occurrences).sum()
            data = float(mean - opt_value)
        return Number(data, "residual_energy")



from __future__ import annotations

import typing as tp
from dataclasses import dataclass

from jijbench.node.base import DataNode
from jijbench.typing import NumberTypes


@dataclass
class Number(DataNode[NumberTypes]):
    """A class representing a numerical element.

    Attributes:
        data (NumberTypes): The int or float object.
        name (str): The name of the element.
    """

    @classmethod
    def validate_data(cls, data: NumberTypes) -> NumberTypes:
        """Validate the data to be a number.

        Args:
            data (NumberTypes): The data to be validated.

        Raises:
            TypeError: If the data is not a number.

        Returns:
            NumberTypes: The validated data.
        """
        return cls._validate_dtype(data, (int, float))


@dataclass
class String(DataNode[str]):
    """A class representing a string element.

    Attributes:
        data (str): The string object.
        name (str): The name of the element.
    """

    @classmethod
    def validate_data(cls, data: str) -> str:
        """Validate the data to be a string.

        Args:
            data (str): The data to be validated.

        Raises:
            TypeError: If the data is not a string.

        Returns:
            str: The validated data.
        """
        return cls._validate_dtype(data, (str,))


@dataclass
class Callable(DataNode[tp.Callable]):
    """A class representing a callable element.

    Attributes:
        data (Callable): The callable object.
        name (str): The name of the element.
    """

    @classmethod
    def validate_data(cls, data: tp.Callable) -> tp.Callable:
        """Validate the data to be a callable.

        Args:
            data (Callable): The data to be validated.

        Raises:
            TypeError: If the data is not a callable.

        Returns:
            Callable: The validated data.

        """
        return cls._validate_dtype(data, (tp.Callable,))



from __future__ import annotations

import re
from numbers import Number
from typing import Callable, Literal, cast

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly
import plotly.express as px
from matplotlib import axes, figure
from matplotlib.ticker import MaxNLocator

import jijbench as jb
from jijbench.exceptions.exceptions import UserFunctionFailedError
from jijbench.visualization.metrics.utils import (
    _create_fig_title_list,
    _df_has_number_array_column_target_name,
    _df_has_number_column_target_name,
    _df_has_valid_multipliers_column,
)

AXIS_LABEL_POS = Literal["top", "bottom"]


def _get_violations_dict(x: pd.Series) -> dict:
    """Get a dictionary of constraint violations from `pd.Series`.

    This function is intended to be used for `jb.Experiment.table` as example.

    Args:
        x (pd.Series): a Series of benchmark result. Expected to include information on constraint violations.

    Returns:
        dict: a dictionary of constraint violations

    Example:
        The code below get the dictionary of constraint violations for each row in experiment.

        ```python
        import jijbench as jb
        import jijzept as jz
        from jijbench.visualization.metrics.plot import _get_violations_dict

        problem = jb.get_problem("TSP")
        instance_data = jb.get_instance_data("TSP")[0][1]
        multipliers1 = {"onehot_time": 0.003, "onehot_location": 0.003}
        multipliers2 = {"onehot_time": 0.3, "onehot_location": 0.3}

        config_path = "XX"
        sa_sampler = jz.JijSASampler(config=config_path)

        bench = jb.Benchmark(
            params = {
                "model": [problem],
                "feed_dict": [instance_data],
                "multipliers": [multipliers1, multipliers2],
            },
            solver = [sa_sampler.sample_model],
        )
        experiment = bench()
        metrics = experiment.table.apply(_get_violations_dict, axis=1)
        ```
    """
    constraint_violations_indices = x.index[x.index.str.contains("violations")]
    return {index: x[index] for index in constraint_violations_indices}


def _calc_mean_of_array(x: pd.Series, column_name: str) -> float:
    """
    Gets the mean of the elements in the specified column ,assuming this element is an np.ndarray.

    This function is intended to be used for `jb.Experiment.table`
    as experiment.table.apply(_calc_mean_of_array, axis=1, column_name=column_name)
    """
    num_occ = x["num_occurrences"]
    array = x[column_name]
    mean = np.sum(num_occ * array) / np.sum(num_occ)
    return mean


def _get_multiplier(x: pd.Series, constraint_name: str) -> float:
    """
    Gets the multiplier of given constraint name from `pd.Series`.

    This function is intended to be used for `jb.Experiment.table`
    as experiment.table.apply(_get_multiplier, axis=1, constraint_name=constraint_name)
    """
    multipliers = x["multipliers"]
    return multipliers[constraint_name]


class MetricsPlot:
    def __init__(self, result: jb.Experiment) -> None:
        """Visualize the metrics of a benchmark result.

         Attributes:
             result (jb.Experiment): a benchmark result.
             parallelplot_axes_list (list[str]): a list of the names of all the metrics that can be plotted by `parallelplot_experiment`.
                 None until `parallelplot_experiment` is called.

        Example:
             Below is the code to boxplot the constraint violations.
             Check the docstring of each method for details.

            ```python
            import jijbench as jb
            import jijzept as jz
            from jijzept.sampler.openjij.sa_cpu import JijSAParameters
            from jijbench.visualization.metrics.plot import MetricsPlot

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers1 = {"onehot_time": 0.003, "onehot_location": 0.003}
            multipliers2 = {"onehot_time": 0.3, "onehot_location": 0.3}

            config_path = "XX"
            sa_parameter = JijSAParameters(num_reads=15)
            sa_sampler = jz.JijSASampler(config=config_path)

            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "parameters": [sa_parameter],
                    "multipliers": [multipliers1, multipliers2],
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()

            mplot = MetricsPlot(result)
            fig_ax_tuple = mplot.boxplot_violations()
        ```
        """
        self.result = result

    def boxplot(
        self,
        f: Callable,
        figsize: tuple[int | float] | None = None,
        title: str | list[str] | None = None,
        title_fontsize: float | None = None,
        xticklabels_size: float | None = None,
        xticklabels_rotation: float | None = None,
        ylabel: str | None = None,
        ylabel_size: float | None = None,
        yticks: list[int | float] | None = None,
        **matplotlib_boxplot_kwargs,
    ):
        """Draw a box and whisker plot of the metrics based on `result` data using matplotlib.boxplot.

        This method applies the function f to the result (i.e. `jb.Experiment`) to get the metrics (pd.Series), and draw boxplot of this metrics.
        This metrics series calculated as metrics = self.result.table.apply(f, axis=1) in this method assumes the following structure.
            The length is equal to the number of rows in `result.table`.
            the element is a dictionary where the key is the name of each boxplot and the value is the np.array of the boxplot data.
        This method returns a figure and axes, so you can post-process them to change the appearance of the plot.
        See also the example below.

        Args:
            f (Callable): Callalbe to apply to table and get metrics.
            figsize (tuple[int | float] | None): the size of figure. The default uses matplotlib's default value.
            title (str | list[str] | None): the title of figure. The default uses the indices of `result.table`.
            title_fontsize (float | None): the fontsize of the title.The default uses matplotlib's default value.
            xticklabels_size (float | None): the fontsize of the xticklabels (i.e. the name of each boxplot). The default uses matplotlib's default value.
            xticklabels_rotation (float | None): the rotation angle of the xticklabels in degree.The default uses matplotlib's default value.
            ylabel (str | None): the ylabel of figure. Defaults to None.
            ylabel_size (float | None): the fontsize of the ylabel. The default uses matplotlib's default value.
            yticks (list[int | float] | None): the yticks of figure. Default to only integers by`MaxNLocator(integer=True)`.
            matplotlib_boxplot_kwargs: the parameter passed to matplotlib.boxplot.

        Returns:
            tuple[tuple[matplotlib.figure.Figure, matplotlib.axes.Subplot]]: A tuple of length equal to the number of rows in result. each element of is a tuple of figure and axes.

        Example:
            The code below draws a boxplot of violations of each constraint.
            Note that result (i.e. `jb.Experiment`) holds violations for each constraint in np.array format.
            In the first example, postprocessing the figure and axes changes the appearance of the plot.

            ```python
            import jijbench as jb
            import jijzept as jz
            from jijzept.sampler.openjij.sa_cpu import JijSAParameters
            from jijbench.visualization.metrics.plot import MetricsPlot
            import pandas as pd

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers1 = {"onehot_time": 0.003, "onehot_location": 0.003}
            multipliers2 = {"onehot_time": 0.3, "onehot_location": 0.3}

            config_path = "XX"
            sa_parameter = JijSAParameters(num_reads=15)
            sa_sampler = jz.JijSASampler(config=config_path)

            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "parameters": [sa_parameter],
                    "multipliers": [multipliers1, multipliers2],
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()

            def get_violations_dict(x: pd.Series) -> dict:
                constraint_violations_indices = x.index[x.index.str.contains("violations")]
                return {index: x[index] for index in constraint_violations_indices}

            mplot = MetricsPlot(result)
            fig_ax_tuple = mplot.boxplot(f=get_violations_dict)

            # you can post-process figure and axes to change the appearance of the plot.
            for fig, ax in fig_ax_tuple:
                fig.suptitle("my title")
                display(fig)
            ```

            By using the `construct_experiment_from_samplesets function`,
            `boxplot` can also be used for `jm.SampleSet` obtained without `JijBenchmark`.

            ```python
            import jijbench as jb
            import jijzept as jz
            from jijbench.visualization.metrics.plot import MetricsPlot
            from jijbench.visualization.metrics.utils import construct_experiment_from_samplesets
            import pandas as pd

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers = {"onehot_time": 0.003, "onehot_location": 0.003}

            config_path = "XX"
            sampler = jz.JijSASampler(config=config_path)
            sampleset = sampler.sample_model(model=problem, feed_dict=instance_data, multipliers=multipliers, num_reads=100)

            def get_violations_dict(x: pd.Series) -> dict:
                constraint_violations_indices = x.index[x.index.str.contains("violations")]
                return {index: x[index] for index in constraint_violations_indices}

            result = construct_experiment_from_samplesets(sampleset)
            mplot = MetricsPlot(result)
            fig_ax_tuple = mplot.boxplot(f=get_violations_dict)
            ```
        """
        metrics = self.result.table.apply(f, axis=1)
        title_list = _create_fig_title_list(metrics, title)

        fig_ax_list = []
        for i, (_, data) in enumerate(metrics.items()):
            fig, ax = plt.subplots(figsize=figsize)
            fig.suptitle(title_list[i], fontsize=title_fontsize)
            ax.boxplot(data.values(), **matplotlib_boxplot_kwargs)
            ax.set_xticklabels(
                data.keys(), size=xticklabels_size, rotation=xticklabels_rotation
            )
            ylabel = cast("str", ylabel)
            ax.set_ylabel(ylabel, size=ylabel_size)
            if yticks is None:
                # make yticks integer only
                ax.yaxis.set_major_locator(MaxNLocator(integer=True))
            else:
                ax.set_yticks(yticks)
            fig_ax_list.append((fig, ax))

        return tuple(fig_ax_list)

    def boxplot_violations(
        self,
        figsize: tuple[int | float] | None = None,
        title: str | list[str] | None = None,
        title_fontsize: float | None = None,
        constraint_name_fontsize: float | None = None,
        constraint_name_fontrotation: float | None = None,
        ylabel: str | None = None,
        ylabel_size: float | None = None,
        yticks: list[int | float] | None = None,
        **matplotlib_boxplot_kwargs,
    ) -> tuple[tuple[figure.Figure, axes.Subplot]]:
        """Draw a box and whisker plot of the constraint violations of `result` data using matplotlib.boxplot.

        The arguments are passed to matplotlib functions to change the appearance of the plot.
        matplotlib_boxplot_kwargs are passed to matplotlib.boxplot, and defaults to `{showmeans: True, whis: [0, 100]}`.
            showmeans=True shows mean values in markers.
            the outliers are not considered and whiskers match maxima and minima by whis=[0, 100].
        This method returns a figure and axes, so you can post-process them to change the appearance of the plot.
        See also the example below.

        Args:
            figsize (tuple[int | float] | None): the size of figure. The default uses matplotlib's default value.
            title (str | list[str] | None): the title of figure. The default uses the indices of `result.table`.
            title_fontsize (float | None): the fontsize of the title.The default uses matplotlib's default value.
            constraint_name_size (float | None): the fontsize of the constraint name (i.e. xticklabels). The default uses matplotlib's default value.
            constraint_name_rotation (float | None): the rotation angle of the constraint name in degree.The default uses matplotlib's default value.
            ylabel (str | None): the ylabel of figure. Defaults to "constraint violations".
            ylabel_size (float | None): the fontsize of the ylabel. The default uses matplotlib's default value.
            yticks (list[int | float] | None): the yticks of figure. Default to only integers by`MaxNLocator(integer=True)`.
            matplotlib_boxplot_kwargs: the parameter passed to matplotlib.boxplot. Defaults to `{showmeans: True, whis: [0, 100]}`.

        Returns:
            tuple[tuple[matplotlib.figure.Figure, matplotlib.axes.Subplot]]: A tuple of length equal to the number of rows in result. each element of is a tuple of figure and axes.


        Example:
            Below is the code to boxplot the constraint violations.
            In the first example, postprocessing the figure and axes changes the appearance of the plot.

            ```python
            import jijbench as jb
            import jijzept as jz
            from jijzept.sampler.openjij.sa_cpu import JijSAParameters
            from jijbench.visualization.metrics.plot import MetricsPlot

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers1 = {"onehot_time": 0.003, "onehot_location": 0.003}
            multipliers2 = {"onehot_time": 0.3, "onehot_location": 0.3}

            config_path = "XX"
            sa_parameter = JijSAParameters(num_reads=15)
            sa_sampler = jz.JijSASampler(config=config_path)

            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "parameters": [sa_parameter],
                    "multipliers": [multipliers1, multipliers2],
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()

            mplot = MetricsPlot(result)
            fig_ax_tuple = mplot.boxplot_violations()

            # you can post-process figure and axes to change the appearance of the plot.
            for fig, ax in fig_ax_tuple:
                fig.suptitle("my title")
                display(fig)
            ```

            By using the `construct_experiment_from_samplesets function`,
            `boxplot_violations` can also be used for `jm.SampleSet` obtained without `JijBenchmark`.

            ```python
            import jijbench as jb
            import jijzept as jz
            from jijbench.visualization.metrics.plot import MetricsPlot
            from jijbench.visualization.metrics.utils import construct_experiment_from_samplesets

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers = {"onehot_time": 0.003, "onehot_location": 0.003}

            config_path = "XX"
            sampler = jz.JijSASampler(config=config_path)
            sampleset = sampler.sample_model(model=problem, feed_dict=instance_data, multipliers=multipliers, search=False, num_reads=100)

            result = construct_experiment_from_samplesets(sampleset)
            mplot = MetricsPlot(result)
            fig_ax_tuple = mplot.boxplot_violations()
            ```
        """
        if ylabel is None:
            ylabel = "constraint violations"
        if len(matplotlib_boxplot_kwargs) == 0:
            # Show the arithmetic means in boxplot.
            matplotlib_boxplot_kwargs["showmeans"] = True
            # Make boxplot whisker positions min and max.
            matplotlib_boxplot_kwargs["whis"] = [0, 100]
        fig_ax_tuple = self.boxplot(
            f=_get_violations_dict,
            figsize=figsize,
            title=title,
            title_fontsize=title_fontsize,
            xticklabels_size=constraint_name_fontsize,
            xticklabels_rotation=constraint_name_fontrotation,
            ylabel=ylabel,
            ylabel_size=ylabel_size,
            yticks=yticks,
            **matplotlib_boxplot_kwargs,
        )

        # Add a horizontal line to indicate that the constraint is satisfied. (violation = 0)
        fig_ax_list = []
        for fig, ax in fig_ax_tuple:
            ax.axhline(0, xmin=0, xmax=1, color="gray", linestyle="dotted")
            fig_ax_list.append((fig, ax))
        return tuple(fig_ax_list)

    def parallelplot_experiment(
        self,
        color_column_name: str | None = None,
        color_midpoint: float | None = None,
        additional_axes: list[str] | None = None,
        additional_axes_created_by_function: dict[str, Callable] | None = None,
        display_axes_list: list[str] | None = None,
        rename_map: dict[str, str] | None = None,
        axis_label_pos: AXIS_LABEL_POS | None = None,
        axis_label_fontsize: Number | None = None,
        title: str | None = None,
        height: Number | None = None,
        width: Number | None = None,
    ) -> plotly.graph_objects.Figure:
        """Plot the parallel plot of the experiment.

        Args:
            color_column_name (str | None): the column name, and the values from this column are used to assign color to mark.
                Defaults to samplemean_total_violations or objective if those columns exist.
            color_midpoint (float | None): the midpoint of the color scale. Defaults to the mean of the color column value.
            additional_axes (list[str] | None): A list of column names for additional axes.
                The conditions for available column names are that they are elements of self.result.table.columns and that the values from the column is number.
                Defaults to None.
            additional_axes_created_by_function (dict[str, Callable]): A list of dict, where the key is the label of the axis and the value is the callable.
                The callable is applied to `self.result.table` as self.result.table.apply(callable, axis=1), and the result is added to axes.
                The callable takes a `pd.Series` and returns a number.
                Defaults to None.
            display_axes_list (list[str] | None): A list of labels for the displayed axes. This argument allows you to select and sort the axes to display.
                Check the `parallelplot_axes_list` attribute for available axes. Defaults to all axes.
            rename_map (dict[str, str] | None): A dictionary where the key is the original axis label and the value is the user-specified axis label.
                Check the original axis labels in the `parallelplot_axes_list` attribute.
                Defaults is None, the original axis labels is displayed.
            axis_label_pos (AXIS_LABEL_POS | None): the position of the axis label. Only "top" or "bottom" are accepted. Defaults to top.
            axis_label_fontsize (Number | None): the fontsize of the axis label. Defaults to None.
            title (str | None): the title of the plot. Defaults to None.
            height (Number | None): the height of the plot. Defaults to None.
            width (Number | None): the width of the plot. Defaults to None.

        Returns:
            plotly.graph_objects.Figure: the parallel plot of the experiment.

        Examples:
            The following example is the most basic usage. A parallel plot of the benchmark results is displayed.

            ```python
            from itertools import product
            import jijbench as jb
            from jijbench.visualization.metrics.plot import MetricsPlot
            import jijzept as jz

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]

            onehot_time_multipliers = [0.01, 0.1]
            onehot_location_multipliers = [0.01, 0.1]
            multipliers = [
                {"onehot_time": onehot_time_multiplier,
                "onehot_location": onehot_location_multiplier}
                for onehot_time_multiplier, onehot_location_multiplier in product(onehot_time_multipliers, onehot_location_multipliers)
            ]
            config_path = XX
            sa_sampler = jz.JijSASampler(config=config_path)
            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "multipliers": multipliers,
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()

            mp = MetricsPlot(result)
            fig = mp.parallelplot_experiment()
            ```

            You can change the appearance of the graph by performing the following operations on the `plotly.graph_objects.Figure` instance returned by `parallelplot_experiment`.
            The example below changes the fontsize of the range.
            For other operations, refer to the plotly reference, https://plotly.com/python/reference/parcoords/.

            ```python
            fig.update_traces(rangefont_size=15, selector=dict(type='parcoords'))
            fig.show()
            ```

            This example gives some arguments to `parallelplot_experiment`.
            `additional_axes` argument adds the execution_time column to the plot. This is the column of result.table and its element are number.
            `additional_axes_created_by_function` argument add the values calculated from result.table to the plot.
            `rename_map` insert line breaks for long data labels to make the charts easier to read. Line breaks are done with <br>.
            `axis_label_pos` is set to bottom to avoid the data label overlapping the figure due to line breaks.

            ```python
            from itertools import product
            import numpy as np
            import pandas as pd

            import jijbench as jb
            from jijbench.visualization.metrics.plot import MetricsPlot
            import jijzept as jz
            from jijzept.sampler.openjij.sa_cpu import JijSAParameters

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]

            onehot_time_multipliers = [0.01, 0.1]
            onehot_location_multipliers = [0.01, 0.1]
            multipliers = [
                {"onehot_time": onehot_time_multiplier,
                "onehot_location": onehot_location_multiplier}
                for onehot_time_multiplier, onehot_location_multiplier in product(onehot_time_multipliers, onehot_location_multipliers)
            ]
            config_path = "XX"
            sa_parameter = JijSAParameters(num_reads=30)
            sa_sampler = jz.JijSASampler(config=config_path)
            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "parameters": [sa_parameter],
                    "multipliers": multipliers,
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()


            def get_num_reads_from_parameters(x: pd.Series) -> float:
                return x["parameters"].num_reads

            def calc_samplemean_energy(x: pd.Series) -> float:
                num_occ	= x["num_occurrences"]
                array = x["energy"]
                mean = np.sum(num_occ * array) / np.sum(num_occ)
                return mean

            mp = MetricsPlot(result)
            fig = mp.parallelplot_experiment(
                additional_axes=["execution_time"],
                additional_axes_created_by_function={
                    "num_reads": get_num_reads_from_parameters,
                    "samplemean_energy": calc_samplemean_energy,
                },
                rename_map={
                    "onehot_time_multiplier": "onehot_time<br>multiplier",
                    "onehot_location_multiplier": "onehot_location<br>multiplier",
                    "samplemean_objective": "samplemean<br>objective",
                    "samplemean_onehot_time_violations": "samplemean<br>onehot_time<br>violations",
                    "samplemean_onehot_location_violations": "samplemean<br>onehot_location<br>violations",
                    "samplemean_total_violations": "samplemean<br>total_violations",
                },
                axis_label_pos="bottom",
            )

            ```

        """
        if additional_axes is None:
            additional_axes = []
        if additional_axes_created_by_function is None:
            additional_axes_created_by_function = {}
        if axis_label_pos is None:
            axis_label_pos = "top"
        if not (axis_label_pos in ["top", "bottom"]):
            raise ValueError(
                f"axis_label_pos must be 'top' or 'bottom', but {axis_label_pos} is given."
            )

        result_table = self.result.table

        # The key is a column name (str), and the value is the data of each column (pd.Series).
        data_to_create_df_parallelplot = {}

        # Displayed data specified by user
        for display_column in additional_axes:
            # Check if the column exists and the elements is number.
            if not _df_has_number_column_target_name(result_table, display_column):
                raise TypeError(
                    f"{display_column}is not a column with number elements."
                )
            data_to_create_df_parallelplot[display_column] = result_table[
                display_column
            ]

        # Data generated by user custom functions
        for column_name, func in additional_axes_created_by_function.items():
            # TODO: デコレータで書き直した方が可読性が上がると思われる。エラーレイズと返り値がnumberであることのチェックをデコレータ内で行う
            try:
                data_to_create_df_parallelplot[column_name] = result_table.apply(
                    func, axis=1
                )
            except Exception as e:
                msg = f'An error occurred inside your function. Please check "{func.__name__}" in additional_axes_created_by_function. -> {e}'
                raise UserFunctionFailedError(msg)
            for value in data_to_create_df_parallelplot[column_name].values:
                if not isinstance(value, Number):
                    raise TypeError(
                        f"{column_name} is not a column with number elements."
                    )

        # multiplires (If self.result has a valid multipliers column)
        if _df_has_valid_multipliers_column(result_table):
            for constraint_name in result_table["multipliers"].values[0].keys():
                data_to_create_df_parallelplot[
                    constraint_name + "_multiplier"
                ] = result_table.apply(
                    _get_multiplier, axis=1, constraint_name=constraint_name
                )

        # num_feasible
        if _df_has_number_column_target_name(result_table, "num_feasible"):
            data_to_create_df_parallelplot["num_feasible"] = result_table[
                "num_feasible"
            ]

        # objective
        if _df_has_number_array_column_target_name(result_table, "objective"):
            data_to_create_df_parallelplot["samplemean_objective"] = result_table.apply(
                _calc_mean_of_array, axis=1, column_name="objective"
            )

        # violations
        for violation_column_name in result_table.columns[
            result_table.columns.str.contains("violations")
        ]:
            if _df_has_number_array_column_target_name(
                result_table, violation_column_name
            ):
                data_to_create_df_parallelplot[
                    "samplemean_" + violation_column_name
                ] = result_table.apply(
                    _calc_mean_of_array,
                    axis=1,
                    column_name=violation_column_name,
                )

        # Extract series about violations from data_to_create_df_parallelplot (key starts with 'samplemean_' and ends with '_violations') and calculates samplemean_total_violations by taking sum.
        start, end = re.compile(r"^samplemean_"), re.compile(r"_violations$")
        violation_series = [
            series
            for name, series in data_to_create_df_parallelplot.items()
            if start.search(name) and end.search(name)
        ]
        if violation_series:
            data_to_create_df_parallelplot["samplemean_total_violations"] = sum(
                violation_series
            )

        self.df_parallelplot = df_parallelplot = pd.DataFrame(
            data_to_create_df_parallelplot
        )

        if display_axes_list is None:
            display_axes_list = self.parallelplot_axes_list
        self.df_parallelplot_displayed = df_parallelplot_displayed = df_parallelplot[
            display_axes_list
        ]

        if color_column_name is None:
            if "samplemean_total_violations" in df_parallelplot_displayed.columns:
                color_column_name = "samplemean_total_violations"
            elif "samplemean_objective" in df_parallelplot_displayed.columns:
                color_column_name = "samplemean_objective"

        if color_midpoint is None and color_column_name is not None:
            color_midpoint = df_parallelplot_displayed[color_column_name].mean()

        fig = px.parallel_coordinates(
            df_parallelplot_displayed.reset_index(drop=True),
            color=color_column_name,
            labels=rename_map,
            color_continuous_scale=px.colors.diverging.Tealrose,
            color_continuous_midpoint=color_midpoint,
            title=title,
            height=height,
            width=width,
        )
        fig.update_traces(labelside=axis_label_pos, selector=dict(type="parcoords"))
        fig.update_traces(
            labelfont_size=axis_label_fontsize, selector=dict(type="parcoords")
        )
        fig.show()

        return fig

    @property
    def parallelplot_axes_list(self) -> list[str]:
        if hasattr(self, "df_parallelplot"):
            return list(self.df_parallelplot.columns)
        else:
            return []



from __future__ import annotations

import abc
import copy
import typing as tp
from dataclasses import dataclass

from jijbench.typing import DataNodeT, DataNodeT2, T


@dataclass
class DataNode(tp.Generic[T], metaclass=abc.ABCMeta):
    """A base class for all data nodes in a computation graph.

    Attributes:
        data: The data in the node.
        name (Hashable): The name of the node, must be hashable.
        operator (FunctionNode | None, optional): The operator that was applied to the node, or None if the node is a leaf.
    """

    data: T
    name: tp.Hashable

    def __post_init__(self) -> None:
        self.operator: FunctionNode | None = None
        setattr(self, "state", None)

    def __setattr__(self, name: str, value: tp.Any) -> None:
        """Set the value of an attribute.

        Args:
            name (str): The name of the attribute.
            value (Any): The new value of the attribute.
        """
        if name == "data":
            value = self.validate_data(value)
        return super().__setattr__(name, value)

    @property
    def dtype(self) -> type:
        """Return the type of the data."""
        return type(self.data)

    @classmethod
    @abc.abstractmethod
    def validate_data(cls, data: T) -> T:
        """Validate the data in the node.
        This method must be implemented by subclasses, and it should raise a TypeError if the
        data is not of a valid type.

        Args:
            data (T): The data to be validated.

        Returns:
            T: The data, if it is valid.
        """
        pass

    @classmethod
    def _validate_dtype(cls, data: T, cls_tuple: tuple) -> T:
        if isinstance(data, cls_tuple):
            return data
        else:
            dtype_str = " or ".join(
                map(
                    lambda x: x.__name__ if hasattr(x, "__name__") else str(x),
                    cls_tuple,
                )
            )
            raise TypeError(
                f"Attribute data of class {cls.__name__} must be type {dtype_str}."
            )

    def _init_attrs(self, node: DataNode) -> None:
        """Refresh the attributes in DataNode object.

        Args:
            node (DataNode): DataNode object.
        """
        operator = node.__dict__.pop("operator")
        state = node.__dict__.pop("state")
        self.__init__(**node.__dict__)
        self.operator = operator
        setattr(self, "state", state)

    def apply(
        self,
        f: FunctionNode[DataNodeT, DataNodeT2],
        others: list[DataNodeT] | None = None,
        **kwargs: tp.Any,
    ) -> DataNodeT2:
        """Apply a function `f` on the data stored in the `DataNode` instance and other input `DataNode` instances.

        Args:
            f (FunctionNode[DataNodeT, DataNodeT2]): The function to be applied on the data.
            others (list[DataNodeT] | None, optional): A list of other `DataNode` instances to be used as inputs to the function. Defaults to None. Defaults to None.
            **kwargs: Arbitrary keyword arguments to be passed to the function.

        Returns:
            DataNodeT2: The result of applying the function on the data.
        """
        inputs = [tp.cast("DataNodeT", copy.copy(self))] + (others if others else [])
        node = f(inputs, **kwargs)
        node.operator = f
        setattr(node, "state", getattr(self, "state"))
        return node


class FunctionNode(tp.Generic[DataNodeT, DataNodeT2], metaclass=abc.ABCMeta):
    """A base class for all function nodes to operate DataNode objects.

    Attributes:
        name (Hashable): The name of the function.
        inputs (list[DataNodeT]): A list of input `DataNode` objects that the function will operate.
    """

    def __init__(self, name: tp.Hashable = None) -> None:
        """Initialize the function node with a name and an empty list of inputs.

        Args:
            name (Hashable, optional): The name of the function. Defaults to None.
        """
        if name is None:
            name = self.__class__.__name__
        self._name = name
        self.inputs: list[DataNodeT] = []

    def __call__(self, inputs: list[DataNodeT], **kwargs: tp.Any) -> DataNodeT2:
        """Operate the inputs to produce a new `DataNode` object.

        Args:
            inputs (list[DataNodeT]): A list of input `DataNode` objects.
            kwargs (Any): Keyword arguments for the operation.

        Returns:
            DataNodeT2: A new `DataNode` object that is the result of the operation.
        """
        node = self.operate(inputs, **kwargs)
        self.inputs += inputs
        return node

    @property
    def name(self) -> tp.Hashable:
        """The name of the function."""
        return self._name

    @name.setter
    def name(self, name: tp.Hashable) -> None:
        """Set the name of the function.

        Args:
            name (Hashable): The new name for the function.

        Raises:
            TypeError: If the specified name is not hashable.
        """
        if not isinstance(name, tp.Hashable):
            raise TypeError(f"{self.__class__.__name__} name must be hashable.")
        self._name = name

    @abc.abstractmethod
    def operate(self, inputs: list[DataNodeT], **kwargs: tp.Any) -> DataNodeT2:
        """Perform the operation on the inputs.
        This method must be implemented by subclasses.

        Args:
            inputs (list[DataNodeT]): A list of input `DataNode` objects.
            kwargs (Any): Keyword arguments for the operation.

        Returns:
            DataNodeT2: A new `DataNode` object that is the result of the operation.
        """
        pass
